﻿// Copyright (c) 2019 by E. Lücke

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices;

namespace G_Api_Wrapper
{
    /// <summary>
    /// This class performs function calls to the g_api_can.dll file.
    /// The library provides access to the CAN hardware functionality.
    /// </summary>
    public class GAPI_CAN
    {

        //**********************************************************************************************************************************************************************
        // CYCLIC MESSAGES
        //**********************************************************************************************************************************************************************

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        // G_Can_CyclicMsgs_DefineMsg
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        // prototype
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        /// <summary>
        /// This command defines the message indicated by id.
        /// </summary>
        /// <param name="portHandle">Handle to the communication port</param>
        /// <param name="id">Identifier of the message to be defined</param>
        /// <remarks>For using extended identifiers or the CAN FD functionality of the interface, you have to call <see cref="G_Can_InitInterface"/> with the command flags set accordingly. 
        /// When sending frames with Extended Identifiers, bit 32 (0x80000000) has to be set in the frame ID value.
        /// When sending CAN FD frames, bit 31 (0x40000000) has to be set in the frame ID value.
        /// </remarks>
        /// <param name="cycleTime">Cycle time in ms</param>
        /// <param name="sendState">Message send state
        /// The following values are possible:
        /// G_CAN__CYCLIC_MSGS__SEND_STATE__DONT_SEND_MSG
        /// G_CAN__CYCLIC_MSGS__SEND_STATE__SEND_MSG</param>
        /// <param name="group">Group ID of the message
        /// In the recent version, only group IDs 0 and 1 are supported.
        /// All messages belonging to the same group can be started at the same time with <see cref="G_Can_CyclicMsgs_StartGroup"/> and stopped with <see cref="G_Can_CyclicMsgs_StopGroup"/>.
        /// If the group ID is 0, the message does not have any group affiliation.</param>
        /// <param name="msgCount">Number of transmit repetitions of the message (1..255; 0=endless)</param>
        /// <param name="data">Pointer to the buffer for data bytes 0..63</param>
        /// <param name="dlc">Data length code (0..15)</param>
        /// <returns>G_NO_ERROR on success, G_ERROR__DLL__CAN__... on error</returns>
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        [DllImport("g_api_can.dll", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 G_Can_CyclicMsgs_DefineMsg(UInt32 portHandle, UInt32 id, UInt16 cycleTime, Byte sendState, Byte group, Byte msgCount, ref Byte data, Byte dlc);
       
        public static UInt32 Call_G_Can_CyclicMsgs_DefineMsg(UInt32 portHandle, UInt32 id, UInt16 cycleTime, Byte sendState, Byte group, Byte msgCount, ref Byte data, Byte dlc)
        {
            return G_Can_CyclicMsgs_DefineMsg(portHandle, id, cycleTime, sendState, group, msgCount, ref data, dlc);
        }


        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        // G_Can_CyclicMsgs_Ramp_Define
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        // parameters
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        /// <list type="bullet">
        /// <listheader>
        /// <term>Ramp parameters structure</term>
        /// </listheader>
        /// <item>
        /// <term>Id</term>
        /// <description>Identifier of the CAN message (0..0x7FF or 0..0x1FFFFFFF)</description>
        /// </item>
        /// <item>
        /// <term>Signal</term>
        /// <description>Signal definition of the ramp</description>
        /// See <see cref="GAPI_Common.G_Common_Signal_t"/>
        /// </item>
        /// <item>
        /// <list type="bullet">
        ///     <listheader>
        ///     <term>Mode</term>
        ///     <description>Ramp mode
        ///     The following values are possible:</description>
        ///     </listheader>
        ///     <item>
        ///     <term>G_CAN__CYCLIC_MSGS__RAMP__MODE__TRIANGULAR</term>
        ///     <description>Triangular ramp</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__CYCLIC_MSGS__RAMP__MODE__SAWTOOTH</term>
        ///     <description>Sawtooth ramp</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__CYCLIC_MSGS__RAMP__MODE__SAWTOOTH_WITH_OVERFLOW_BIT</term>
        ///     <description>Sawtooth ramp with overflow bit</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__CYCLIC_MSGS__RAMP__MODE__STOP</term>
        ///     <description>The ramp is stopped</description>
        ///     </item>
        ///     </list>
        /// </item>
        /// <item>
        /// <list type="bullet">
        ///     <listheader>
        ///     <term>Direction</term>
        ///     <description>Starting direction
        ///     The following values are possible:</description>
        ///     </listheader>
        ///     <item>
        ///     <term>G_CAN__CYCLIC_MSGS__RAMP__DIRECTION__DOWN</term>
        ///     <description>Ramp starts downwards</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__CYCLIC_MSGS__RAMP__DIRECTION__UP</term>
        ///     <description>Ramp starts upwards</description>
        ///     </item>
        ///     </list>
        /// </item>
        /// <item>
        /// <term>MinValue</term>
        /// <description>Minimum value of the ramp</description>
        /// </item>
        /// <item>
        /// <term>MaxValue</term>
        /// <description>Maximum value of the ramp</description>
        /// </item>
        /// <item>
        /// <term>UpdateRate</term>
        /// <description>Updating the ramp
        /// 1: with each CAN message
        /// 2: with every second CAN message
        /// etc.</description>
        /// </item>
        /// <item>
        /// <term>Increment</term>
        /// <description>Increment of the ramp</description>
        /// </item>
        /// <item>
        /// <term>RampSum</term>
        /// <description>Relative ramp sum (sum of the partial increments starting with the execution of this command, therefore relative)
        /// (0 = No limitation)</description>
        /// </item>
        /// <item>
        /// <term>NumberOfFlanks</term>
        /// <description>Number of flanks to be used
        /// (0 = No limitation)</description>
        /// </item>
        /// <item>
        /// <term>reserved1</term>
        /// <description>reserved parameter (must be initialized with 0)</description>
        /// </item>
        /// <item>
        /// <term>reserved2</term>
        /// <description>reserved parameter (must be initialized with 0)</description>
        /// </item>
        /// </list>
        /// <remarks>The ramp starts using the current signal value. To avoid unexpected behavior, it is recommended to ensure that the current signal value is within a valid range.
        /// NumberOfFlanks and RampSum are abort conditions. When both values do equal 0, the ramp runs endlessly, otherwise the ramp stops when meeting one of these conditions.
        /// When a ramp runs, the partial increments are added continuously with a stepwidth defined with Increment.</remarks>

        [StructLayout(LayoutKind.Sequential)]
        public struct G_Can_CyclicMsgs_Ramp_Define_Parameters_t
        {
            public UInt32 Id;
            public GAPI_Common.G_Common_Signal_t Signal;
            public Byte Mode;
            public Byte Direction;
            public UInt16 MinValue;
            public UInt16 MaxValue;
            public UInt16 UpdateRate;
            public Double Increment;
            public UInt32 RampSum;
            public UInt16 NumberOfFlanks;
            public Byte reserved1;
            public Byte reserved2;
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        // prototype
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        /// <summary>
        /// This command defines and starts a new ramp or changes the mode of a running ramp within a CAN message.
        /// </summary>
        /// <remarks>A ramp serves to create alternately increasing / decreasing signal courses.</remarks>
        /// <param name="portHandle">Handle to the communication port</param>
        /// <param name="parameters">Ramp parameters. See <see cref="G_Can_CyclicMsgs_Ramp_Define_Parameters_t"/></param>
        /// <returns>G_NO_ERROR on success, G_ERROR__DLL__CAN__... on error</returns>

        [DllImport("g_api_can.dll", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 G_Can_CyclicMsgs_Ramp_Define(UInt32 portHandle, ref G_Can_CyclicMsgs_Ramp_Define_Parameters_t parameters);

        public static UInt32 Call_G_Can_CyclicMsgs_Ramp_Define(UInt32 portHandle, ref G_Can_CyclicMsgs_Ramp_Define_Parameters_t parameters)
        {
            return G_Can_CyclicMsgs_Ramp_Define(portHandle, ref parameters);
        }


        //**********************************************************************************************************************************************************************
        // MONITOR
        //**********************************************************************************************************************************************************************

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        // G_Can_Monitor_BufferMode_Start
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        // prototype
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        /// <summary>
        /// This command starts the monitor in the “BufferMode”.
        /// </summary>
        /// <remarks>In the mode “BufferMode”, the messages come in in succession into an internal ring buffer after passing the monitor filter as sent on the bus/ received by the bus. 
        /// The size of this internal ring buffer is given by bufferSize.</remarks>
        /// <param name="portHandle">Handle to the communication port</param>
        /// <param name="mode">Variable of type G_Can_Monitor_BufferMode_Mode_t indicating the messages coming in into the ring buffer</param>
        /// <remarks>
        /// <list type="bullet">
        ///     <listheader>
        ///     <term>The following values are available:</term>
        ///     </listheader>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__NOTHING</term>
        ///     <description>No messages</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__RX</term>
        ///     <description>RX (Received messages)</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__TX</term>
        ///     <description>TX (Transmitted messages)</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__TX_AND_RX</term>
        ///     <description>RX + TX (Received and transmitted messages)</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__ERROR_FRAME</term>
        ///     <description>Error frames</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__ERROR_FRAME_AND_RX</term>
        ///     <description>Error frames + RX</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__ERROR_FRAME_AND_TX</term>
        ///     <description>Error frames + TX</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_MODE__MODE__ERROR_FRAME_AND_TX_AND_RX</term>
        ///     <description>Error frames + TX + RX</description>
        ///     </item>
        /// </list>
        /// </remarks>
        /// <param name="bufferSize">Buffer size of the ring buffer in bytes (e.g. 20000), or 0 for default size</param>
        /// <returns>G_NO_ERROR on success, G_ERROR__DLL__CAN__... on error</returns>

        [DllImport("g_api_can.dll", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 G_Can_Monitor_BufferMode_Start(UInt32 portHandle, Byte mode, UInt32 bufferSize);

        public static UInt32 Call_G_Can_Monitor_BufferMode_Start(UInt32 portHandle, Byte mode, UInt32 bufferSize)
        {
            return G_Can_Monitor_BufferMode_Start(portHandle, mode, bufferSize);
        }


        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        // G_Can_Monitor_BufferMode_Stop
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        // prototype
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        /// <summary>
        /// This command stops the monitor in the “BufferMode”.
        /// </summary>
        /// <param name="portHandle">Handle to the communication port</param>
        /// <returns>G_NO_ERROR on success, G_ERROR__DLL__CAN__... on error</returns>

        [DllImport("g_api_can.dll", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 G_Can_Monitor_BufferMode_Stop(UInt32 portHandle);

        public static UInt32 Call_G_Can_Monitor_BufferMode_Stop(UInt32 portHandle)
        {
            return G_Can_Monitor_BufferMode_Stop(portHandle);
        }


        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        // G_Can_Monitor_BufferMode_GetItems
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        // parameters
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        /// <list type="bullet">
        /// <listheader>
        /// <term>BufferItem parameters</term>
        /// </listheader>
        /// <item>
        /// <term>Timestamp</term>
        /// <description>Time stamp in ns</description>
        /// </item>
        /// <item>
        /// <term>Id</term>
        /// <description>Identifier</description>
        /// </item>
        /// <item>
        /// <list type="bullet">
        ///     <listheader>
        ///     <term>Flags</term>
        ///     <description>Variable of type G_Can_Monitor_BufferItem_Flags_t
        ///     The following return flags are available:</description>
        ///     </listheader>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_ITEM__FLAG__NONE</term>
        ///     <description>No flag was set</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_ITEM__FLAG__EXTENDED_ID</term>
        ///     <description>29 bit identifiers</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_ITEM__FLAG__TRANSMIT</term>
        ///     <description>Transmitted message (TX)</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_ITEM__FLAG__ERROR_FRAME</term>
        ///     <description>Error frame</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_ITEM__FLAG__EVENT</term>
        ///     <description>Event</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_ITEM__FLAG__BUFFER_OVERRUN</term>
        ///     <description>Monitor buffer overrun of the firmware</description>
        ///     </item>
        ///     <item>
        ///     <term>G_CAN__MONITOR__BUFFER_ITEM__FLAG__DLL_BUFFER_OVERRUN</term>
        ///     <description>Monitor buffer overrun in the API
        ///     If this flag is set, the buffer size given by <see cref="G_Can_Monitor_BufferMode_Start"/> was too small for the incoming entries.</description>
        ///     </item>
        ///     </list>
        /// </item>
        /// <item>
        /// <term>Dlc</term>
        /// <description>Data length code (0..15)</description>
        /// </item>
        /// <item>
        /// <term>DataLength</term>
        /// <description>Data length (0..8)</description>
        /// </item>
        /// <item>
        /// <term>Data</term>
        /// <description>Buffer with data bytes - sized of 8 bytes</description>
        /// </item>
        /// </list>

        [StructLayout(LayoutKind.Sequential)]
        public struct G_Can_Monitor_BufferItem_t
        {
            public UInt64 Timestamp;
            public UInt32 Id;
            public UInt16 Flags;
            public Byte Dlc;
            public Byte DataLength;

            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
            public Byte[] Data;

            public override string ToString()
            {
                String ret;

                ret = String.Format("{0} {1:X} [", Timestamp, Id);

                for (int i = 0; i < Data.Length; i++)
                {
                    ret = ret + " " + String.Format("{0:X2}", Data[i]);
                }

                ret = ret + "]";

                return ret;
            }
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------
        // prototype
        //----------------------------------------------------------------------------------------------------------------------------------------------------------------------

        /// <summary>
        /// This command requests the monitor buffer entries.
        /// </summary>
        /// <param name="portHandle">Handle to the communication port</param>
        /// <param name="bufferSize">Size of bufferItems</param>
        /// <param name="numberOfItems">Returns the number of items returned in bufferItems</param>
        /// <param name="bufferItems">Returns buffer items, see <see cref="G_Can_Monitor_BufferItem_t"/></param>
        /// <returns>G_NO_ERROR on success, G_ERROR__DLL__CAN__... on error</returns>

        [DllImport("g_api_can.dll", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 G_Can_Monitor_BufferMode_GetItems(UInt32 portHandle, UInt32 bufferSize, ref UInt32 numberOfItems, ref Byte bufferItems);

        public static UInt32 Call_G_Can_Monitor_BufferMode_GetItems(UInt32 portHandle, UInt32 bufferSize, ref UInt32 numberOfItems, ref G_Can_Monitor_BufferItem_t[] bufferItems)
        {
            // create buffer for the data
            Byte[] data = new Byte[bufferSize];

            // call API to fill byte array
            UInt32 returnValue = G_Can_Monitor_BufferMode_GetItems(portHandle, bufferSize, ref numberOfItems, ref data[0]);

            // error request
            if (GAPI.G_NO_ERROR == returnValue)
            {
                // no error, compilation from byte array to a structure array
                bufferItems = GAPI_Helpers.ByteArrayToStructureArray<G_Can_Monitor_BufferItem_t>(data, numberOfItems, 24);
             }
            else
            {
                // error, so empty array and return
                bufferItems = new G_Can_Monitor_BufferItem_t[0];
            }

            // return result from API call as result from call function
            return returnValue;
        }

    }
}
